/*
 * deviced
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdbool.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <stdlib.h>
#include <sys/types.h>
#include <vconf.h>

#include "hbm.h"
#include "lbm.h"
#include "auto-brightness-sensorhub.h"
#include "display/util.h"
#include "display/core.h"
#include "display/display-ops.h"
#include "core/device-notifier.h"
#include "battery/power-supply.h"

#define LOWBATCAPACITY 5

static int auto_brightness_state = SETTING_BRIGHTNESS_AUTOMATIC_OFF;

static void change_brightness_transit(int start, int end)
{
	backlight_ops.transit_brt(start, end,
			display_conf.brightness_change_step);
}

static void set_brightness_level(int level)
{
	int old_level = 0;
	if (pm_cur_state != S_NORMAL)
		return;
/*
	if (battery.capacity <= LOWBATCAPACITY && battery.charge_now != CHARGER_CHARGING)
		return;
*/
	_I("level changed! %d", level);

	if ((level > PM_MAX_BRIGHTNESS || level < PM_MIN_BRIGHTNESS) &&
	    (level != LBM_LEVEL && level != HBM_LEVEL && level != SPECIFIC_MODE_OFF)) {
		_E("Invalid level %d", level);
		return;
	}

	backlight_ops.get_brightness(&old_level);

	switch (level) {
	case LBM_LEVEL:
		/*
		 * received LBM enable
		 * change brightness to under LBM_BRIGHTNESS_LOW and set lbm state
		 */
		if (hbm_get_state()) {
			hbm_set_state(false);
			_I("HBM mode disabled.");
		}
		if (!lbm_get_state()) {
			lbm_set_state(true);
			_I("LBM mode enabled.");
		}
		break;
	case HBM_LEVEL:
		/* HBM must do not turn on at DIM_STY state. */
		if (pm_status_flag & DIM_MASK)
			break;

		/*
		 * received HBM enable
		 * change brightness to MAX value and set hbm state
		 */
		if (lbm_get_state()) {
			lbm_set_state(false);
			_I("Disabling LBM before enabling HBM.");
		}

		if (!hbm_get_state()) {
			change_brightness_transit(old_level, PM_MAX_BRIGHTNESS);
			hbm_set_state(true);
			_I("HBM mode enabled.");
		}

		break;
	case SPECIFIC_MODE_OFF:
		/*
		 * received HBM disable
		 * unset hbm state and change brightness to setting value
		 */

		level = backlight_ops.get_default_brt();

		if (hbm_get_state())
			hbm_set_state(false);

		if (lbm_get_state())
			lbm_set_state(false);

		if (old_level != level)
			change_brightness_transit(old_level, level);

		break;
	default:
		/*
		 * received automatic brightness by the light sensor
		 */
		if (auto_brightness_state != SETTING_BRIGHTNESS_AUTOMATIC_ON)
			return;

		if (hbm_get_state())
			hbm_set_state(false);

		if (lbm_get_state())
			lbm_set_state(false);

		if (old_level != level) {
			change_brightness_transit(old_level, level);
			backlight_ops.set_default_brt(level);
		}

		break;
	}

        device_notify(DEVICE_NOTIFIER_LCD_AUTOBRT_SENSING, NULL);
}

static void set_default_brightness(void)
{
	int default_brt;

	default_brt = backlight_ops.get_default_brt();
	backlight_ops.set_default_brt(default_brt);
	backlight_ops.update();
}

static void set_automatic_state_cb(keynode_t *key_nodes, void *data)
{
	if (key_nodes == NULL) {
		_E("Wrong parameter, key_nodes is null.");
		return;
	}
	auto_brightness_state = vconf_keynode_get_int(key_nodes);

	switch (auto_brightness_state) {
	case SETTING_BRIGHTNESS_AUTOMATIC_OFF:
		if (hbm_get_state())
			hbm_set_state(false);
		/* escape dim state if it's in low battery.*/
		set_brightness_changed_state();
		set_default_brightness();
		break;
	case SETTING_BRIGHTNESS_AUTOMATIC_PAUSE:
		if (hbm_get_state())
			hbm_set_state(false);
		break;
	case SETTING_BRIGHTNESS_AUTOMATIC_ON:
		set_brightness_changed_state();
		break;
	default:
		_E("Invalid value %d.", auto_brightness_state);
	}
}

int prepare_level_handler(void)
{
	int status, ret;

	display_info.set_brightness_level = set_brightness_level;

	ret = vconf_get_int(VCONFKEY_SETAPPL_BRIGHTNESS_AUTOMATIC_INT, &status);
	if (ret >= 0)
		auto_brightness_state = status;

	vconf_notify_key_changed(VCONFKEY_SETAPPL_BRIGHTNESS_AUTOMATIC_INT,
				 set_automatic_state_cb, NULL);

	return 0;
}

void exit_level_handler(void)
{
	vconf_ignore_key_changed(VCONFKEY_SETAPPL_BRIGHTNESS_AUTOMATIC_INT,
				 set_automatic_state_cb);
	set_default_brightness();
}
