/*
 * deviced
 *
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <error.h>
#include <device-node.h>

#include "auto-brightness-sensorhub.h"
#include "core/device-notifier.h"
#include "display/util.h"
#include "display/core.h"
#include "display/poll.h"
#include "core/common.h"
#include "core/devices.h"
#include "display/display-actor.h"
#include "display/display-ops.h"
#include "display-info.h"

#define CHARGER_LCD_NODE "/sys/class/power_supply/battery/lcd"
#define SIGNAL_HOMESCREEN       "HomeScreen"
#define CLOCK_CHANGED           "clockchanged"
#define CLOCK_END				"clockend"

enum charging_lcd_state {
	CHARGING_LCD_OFF = 0,
	CHARGING_LCD_ON = 1,
};

static guint autobrt_timer;
static bool lcdon_from_aod;
static int autobrtlevel;

static gboolean display_autobrt_changed(gpointer data)
{

	int level = (int) data;

	autobrt_timer = 0;
	display_info.set_brightness_level(level);

	return G_SOURCE_REMOVE;
}

static GVariant *dbus_autobrightnesschanged(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret = 0;
	int level;

	g_variant_get(param, "(i)", &level);

	/* If received level is identical to autobrt timer reserved level, ignore it. */
	if (autobrt_timer && autobrtlevel == level) {
		_D("Ignore autobrightness change request as same request(%d) is pending.", level);
		return g_variant_new("(i)", ret);
	}

	if (display_info.set_brightness_level) {
		if (autobrt_timer)
			g_source_remove(autobrt_timer);

		/* When display state changes from AOD to LCDON, brightness change
		 * effect seems sluggish because of heavy load of jobs for turning on
		 * display. So delay this brightness change a bit to avoid this
		 * heavy loaded time and therefore make it change smoothly. */
		if (lcdon_from_aod) {
			autobrtlevel = level;
			autobrt_timer = g_timeout_add(200, display_autobrt_changed, (gpointer)level);
		} else {
			display_autobrt_changed((gpointer)level);
		}
	} else {
		ret = -ENOTSUP;
	}

	return g_variant_new("(i)", ret);
}

static void aod_change_signal(GDBusConnection  *conn,
							const gchar      *sender,
							const gchar      *path,
							const gchar      *iface,
							const gchar      *name,
							GVariant         *param,
							gpointer          data)

{
	char *screen = NULL;

	if (!dh_get_param_from_var(param, "(s)", &screen)) {
		_E("failed to get params from gvariant. expected:%s, type:%s", "(s)", g_variant_get_type_string(param));
		goto out;
	}

	if (screen == NULL) {
		_E("screen is null.");
		goto out;
	}

	/* clock-viewer sends "clockchanged" as signal parameter when AOD->LCDON.
	 * On catching this signal, do change brightness immediately. */
	if (!strcmp(screen, CLOCK_CHANGED)) {
		lcdon_from_aod = false;
		if (autobrt_timer) {
			g_source_remove(autobrt_timer);
			display_autobrt_changed((gpointer)autobrtlevel);
		}
	} else if (!strcmp(screen, CLOCK_END)) {
		lcdon_from_aod = true;
	}

out:
	if (screen)
		g_free(screen);
}

static const dbus_method_s dbus_methods[] = {
	{ "AutoBrightnessChanged",   "i",   "i", dbus_autobrightnesschanged },
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_DISPLAY,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static int display_state_changed(void *data)
{
	enum state_t state;
	int ret = 0;

	state = DATA_VALUE_INT(data);

	if (state == S_LCDON)
		ret = sys_set_int(CHARGER_LCD_NODE, CHARGING_LCD_ON);
	else
		ret = sys_set_int(CHARGER_LCD_NODE, CHARGING_LCD_OFF);

	if (ret < 0)
		_E("Can't write %s node %d.", CHARGER_LCD_NODE, state);

	return 0;
}

static void display_handler_init(void *data)
{
	int ret;

	prepare_level_handler();
	register_notifier(DEVICE_NOTIFIER_LCD, display_state_changed);

	lcdon_from_aod = false;
	ret = subscribe_dbus_signal(NULL,
				DEVICED_OBJECT_PATH,
				DEVICED_INTERFACE_NAME,
				SIGNAL_HOMESCREEN,
				aod_change_signal,
				NULL, NULL);
	if (ret <= 0)
		_E("Failed to register signal handler: %d", ret);

	ret = sys_set_int(CHARGER_LCD_NODE, CHARGING_LCD_ON);
	if (ret < 0)
		_E("Can't write %s node.", CHARGER_LCD_NODE);

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_DISPLAY, &dbus_interface);
	if (ret < 0)
		_E("Failed to register dbus object.");
}

static const struct display_ops display_handler_ops = {
	.name     = "dislay-handler",
	.init     = display_handler_init,
};

DISPLAY_OPS_REGISTER(&display_handler_ops)
