/*
 * deviced
 *
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdbool.h>
#include <device-node.h>
#include <vconf.h>
#include <hw/display.h>
#include <libsyscommon/dbus-system.h>

#include "display/core.h"
#include "display/display-ops.h"
#include "core/devices.h"
#include "core/log.h"
#include "core/common.h"
#include "core/device-notifier.h"

#define VCONF_HIGH_CONTRAST		VCONFKEY_SETAPPL_ACCESSIBILITY_HIGH_CONTRAST
#define VCONF_GREYSCALE			VCONFKEY_SETAPPL_ACCESSIBILITY_GREYSCALE

static bool negative_status;
static bool greyscale_status;
static bool greysacle_setting;

static int enhance_update_state(void)
{
	int state;

	_I("Greyscale %d negative %d", greyscale_status, negative_status);

	if (greyscale_status) {
		if (negative_status)
			state = DISPLAY_IMAGE_EFFECT_GREY_NEGATIVE;
		else
			state = DISPLAY_IMAGE_EFFECT_GREY;
	} else if (negative_status)
		state = DISPLAY_IMAGE_EFFECT_NEGATIVE;
	else
		state = DISPLAY_IMAGE_EFFECT_STANDARD;

	backlight_ops.set_image_effect(state);

	return state;
}

static void enhance_greyscale_cb(keynode_t *in_key, void *data)
{
	int val, ret;

	ret = vconf_get_bool(VCONF_GREYSCALE, &val);
	if (ret < 0) {
		_E("Failed to get vconf value for greyscale: %d", vconf_get_ext_errno());
		return;
	}

	_I("Set greyscale scenario (%d)", val);
	greyscale_status = val;
	greysacle_setting = val;

	enhance_update_state();
}

static void enhance_high_contrast_cb(keynode_t *in_key, void *data)
{
	int val, ret;

	ret = vconf_get_bool(VCONF_HIGH_CONTRAST, &val);
	if (ret < 0) {
		_E("Failed to get vconf value for high contrast: %d", vconf_get_ext_errno());
		return;
	}

	_I("Set negative scenario (%d)", val);
	negative_status = val;

	enhance_update_state();
}

static int enhance_lcd_state_changed(void *data)
{
	int state;

	state = DATA_VALUE_INT(data);

	if (state == S_NORMAL)
		enhance_update_state();

	return 0;
}

static void enhance_init(void *data)
{
	int val, ret;
	int state;

	/* register notifier */
	register_notifier(DEVICE_NOTIFIER_LCD, enhance_lcd_state_changed);

	ret = vconf_notify_key_changed(VCONF_HIGH_CONTRAST,
	    enhance_high_contrast_cb, NULL);
	if (ret < 0)
		_E("Failed to register notify: %s", VCONF_HIGH_CONTRAST);

	ret = vconf_get_bool(VCONF_HIGH_CONTRAST, &val);
	if (ret < 0)
		_E("Failed to get vconf value for high contrast: %d", vconf_get_ext_errno());

	/* update negative status */
	_I("Set high contrast (%d)", val);
	negative_status = val;

	ret = vconf_notify_key_changed(VCONF_GREYSCALE,
	    enhance_greyscale_cb, NULL);
	if (ret < 0)
		_E("Failed to register notify: %s", VCONF_GREYSCALE);

	ret = vconf_get_bool(VCONF_GREYSCALE, &val);
	if (ret < 0) {
		_E("Failed to get vconf value for greyscale: %d", vconf_get_ext_errno());
		return;
	}
	greyscale_status = val;
	greysacle_setting = val;

	state = enhance_update_state();

	ret = dbus_handle_emit_dbus_signal(NULL,
					DEVICED_PATH_DISPLAY,
					DEVICED_INTERFACE_DISPLAY,
					"Enhance",
					g_variant_new("(i)", state));
	if (ret < 0)
		_E("Failed to send dbus signal.");
}

static void enhance_exit(void *data)
{
	/* unregister notifier */
	unregister_notifier(DEVICE_NOTIFIER_LCD, enhance_lcd_state_changed);

	vconf_ignore_key_changed(VCONF_HIGH_CONTRAST, enhance_high_contrast_cb);
	vconf_ignore_key_changed(VCONF_GREYSCALE, enhance_greyscale_cb);
}

static const struct display_ops display_enhance_ops = {
	.name     = "enhance",
	.init     = enhance_init,
	.exit     = enhance_exit,
};

DISPLAY_OPS_REGISTER(&display_enhance_ops)
