/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdbool.h>
#include <assert.h>
#include <limits.h>
#include <vconf.h>
#include <fcntl.h>
#include <bundle.h>
#include <eventsystem.h>
#include <hw/battery.h>
#include <libsyscommon/dbus-system.h>

#include "lowbat-handler.h"
#include "battery-ops.h"
#include "battery.h"
#include "config.h"
#include "core/log.h"
#include "core/launch.h"
#include "core/devices.h"
#include "core/device-notifier.h"
#include "core/common.h"
#include "core/list.h"
#include "core/udev.h"
#include "shared/eventsystem.h"
#include "device-node.h"
#include "display/setting.h"
#include "display/poll.h"
#include "display/display-ops.h"
#include "power/power-handler.h"
#include "apps/apps.h"
#include "power-supply.h"

#define BATTERY_UNKNOWN		-1
#define BATTERY_DISABLED	0
#define BATTERY_ENABLED		1

#define VCONF_KEY_BATTERY_WARNING_LEVEL "db/sysman/battery_warning_level"

#define MIN_INOW_VALUE           -50

enum custom_level_control_type {
	CUSTOM_CONTROL_NONE,
	LOWER_THAN_WARNING,
	HIGHER_THAN_WARNING,
};

struct lowbat_process_entry {
	int old;
	int now;
	int (*func) (void *data);
};

static int cur_bat_state = BATTERY_UNKNOWN;
static int cur_bat_capacity = BATTERY_UNKNOWN;
static int custom_warning_level = BATTERY_UNKNOWN;
static bool custom_warning_launch;

struct battery_config_info battery_info = {
	.normal   = BATTERY_NORMAL,
	.warning  = BATTERY_WARNING,
	.critical = BATTERY_CRITICAL,
	.poweroff = BATTERY_POWEROFF,
	.realoff  = BATTERY_REALOFF,
};

static dd_list *lpe;
static int scenario_count;

static guint low_batt_sig_timer;

static int lowbat_monitor_init(void *data);

static int lowbat_initialized(void *data)
{
	static int status = BATTERY_UNKNOWN;

	if (!data)
		return status;

	status = *(int *)data;
	return status;
}

static int lowbat_scenario(int old, int now, void *data)
{
	dd_list *n;
	struct lowbat_process_entry *scenario;
	int found = 0;

	if (old == now && battery.charge_now)
		return found;
	DD_LIST_FOREACH(lpe, n, scenario) {
		if (old != scenario->old || now != scenario->now)
			continue;
		if (!scenario->func)
			continue;
		scenario->func(data);
			found = 1;
		break;
	}
	return found;
}

static int lowbat_add_scenario(int old, int now, int (*func)(void *data))
{
	struct lowbat_process_entry *scenario;

	_D("Lowbat_add_scenario: old(%d) now(%d) func(%p)", old, now, func);

	if (!func) {
		_E("Invalid func address.");
		return -EINVAL;
	}

	scenario = malloc(sizeof(struct lowbat_process_entry));
	if (!scenario) {
		_E("Failed to malloc for notifier.");
		return -ENOMEM;
	}

	scenario->old = old;
	scenario->now = now;
	scenario->func = func;

	DD_LIST_APPEND(lpe, scenario);
	scenario_count++;
	return 0;
}

static int power_execute(void *data)
{
	static const struct device_ops *ops;

	FIND_DEVICE_INT(ops, POWER_OPS_NAME);

	return ops->execute(data);
}

static int booting_done(void *data)
{
	int status;
	static int done;
	static int popup;

	if (data == NULL) {
		if (!done)
			popup = 1;
		goto out;
	}

	status = lowbat_initialized(NULL);
	if (status == BATTERY_UNKNOWN)
		lowbat_monitor_init(NULL);

	done = *(int *)data;
	if (!done)
		goto out;
	_I("Booting done.");
	if (popup) {
		popup = 0;
		if (battery.charge_now != CHARGER_CHARGING &&
		    battery.charge_full != CHARGING_FULL)
			lowbat_popup(BAT_OPT_NONE);
		else
			_I("Skip low battery popup during charging.");
	}
out:
	return done;
}

int lowbat_popup(int option)
{
	static int launched_poweroff;
	static int lowbat_popup_option;
	static char *value;

	if (option == BAT_OPT_NONE) {
		if (!value)
			return -1;
		else
			goto direct_launch;
	}

	if (option == BAT_OPT_POWEROFF)
		launched_poweroff = 0;

	switch (option) {
	case BAT_OPT_CRITICAL:
		value = "lowbattery_critical";
		break;
	case BAT_OPT_WARNING:
		value = "lowbattery_warning";
		break;
	case BAT_OPT_POWEROFF:
		value = "poweroff";
		break;
	case BAT_OPT_ERR_TEMP_LOW:
		value = "chargeerrlow";
		break;
	case BAT_OPT_ERR_TEMP_HIGH:
		value = "chargeerrhigh";
		break;
	case BAT_OPT_ERR_CF_OPEN:
		value = "battdisconnect";
		break;
	default:
		return -1;
	}

	lowbat_popup_option = option;

direct_launch:
	_D("Popup value=%s", value);
	if (booting_done(NULL)) {

		if (launched_poweroff == 1) {
			_I("Will be foreced power off.");
			power_execute(POWER_POWEROFF);
			return 0;
		}

		if (lowbat_popup_option == BAT_OPT_POWEROFF)
			launched_poweroff = 1;

		if (battery_do_not_disturb()) {
			_I("block LCD and %s Popup", value);
			return 0;
		}
		if (lowbat_popup_option == BAT_OPT_ERR_TEMP_LOW ||
		    lowbat_popup_option == BAT_OPT_ERR_TEMP_HIGH ||
		    lowbat_popup_option == BAT_OPT_ERR_CF_OPEN) {
			if (disp_plgn.pm_change_internal)
				disp_plgn.pm_change_internal(INTERNAL_LOCK_LOWBAT, LCD_DIM);
		} else {
			if (disp_plgn.pm_change_internal)
				disp_plgn.pm_change_internal(INTERNAL_LOCK_LOWBAT, LCD_NORMAL);
		}
		if (lowbat_popup_option == BAT_OPT_ERR_TEMP_LOW ||
		    lowbat_popup_option == BAT_OPT_ERR_TEMP_HIGH ||
		    lowbat_popup_option == BAT_OPT_ERR_CF_OPEN)
			return launch_system_app(APP_ABNORMAL,
				2, APP_KEY_TYPE, value);
		return launch_system_app(APP_DEFAULT,
				2, APP_KEY_TYPE, value);
	} else
		_D("Boot-animation running yet.");

	return 0;
}

static int battery_warning_low_act(void *data)
{
	if (battery_plgn.lowbat_noti_launch)
		battery_plgn.lowbat_noti_launch(cur_bat_capacity, BAT_OPT_WARNING);

	(void)lowbat_popup(BAT_OPT_WARNING);
	return 0;
}

static int battery_critical_low_act(void *data)
{
	if (battery_plgn.lowbat_noti_launch)
		battery_plgn.lowbat_noti_launch(cur_bat_capacity, BAT_OPT_CRITICAL);

	(void)lowbat_popup(BAT_OPT_CRITICAL);
	return 0;
}

int battery_power_off_act(void *data)
{
	CRITICAL_LOG("Low battery power off.");
	return power_execute(POWER_POWEROFF);
}

int battery_charge_err_cf_act(void *data)
{
	return lowbat_popup(BAT_OPT_ERR_CF_OPEN);
}

int battery_charge_err_low_act(void *data)
{
	disp_plgn.pm_lock_internal(INTERNAL_LOCK_OVERCOOL, LCD_OFF, STAY_CUR_STATE, 60000);
	return lowbat_popup(BAT_OPT_ERR_TEMP_LOW);
}

int battery_charge_err_high_act(void *data)
{
	disp_plgn.pm_lock_internal(INTERNAL_LOCK_OVERHEAT, LCD_OFF, STAY_CUR_STATE, 60000);
	return lowbat_popup(BAT_OPT_ERR_TEMP_HIGH);
}

static void lowbat_scenario_init(void)
{
	int ret;

	ret = vconf_get_int(VCONF_KEY_BATTERY_WARNING_LEVEL, &custom_warning_level);
	if (ret < 0)
		_E("Failed to get vconf value for battery warning level: %d", vconf_get_ext_errno());

	if (custom_warning_level != BATTERY_UNKNOWN)
		battery_info.warning = custom_warning_level;

	lowbat_add_scenario(battery_info.normal, battery_info.warning, battery_warning_low_act);
	lowbat_add_scenario(battery_info.normal, battery_info.critical, battery_critical_low_act);
	lowbat_add_scenario(battery_info.normal, battery_info.poweroff, battery_critical_low_act);
	lowbat_add_scenario(battery_info.normal, battery_info.realoff, battery_power_off_act);
	lowbat_add_scenario(battery_info.warning, battery_info.warning, battery_warning_low_act);
	lowbat_add_scenario(battery_info.warning, battery_info.critical, battery_critical_low_act);
	lowbat_add_scenario(battery_info.warning, battery_info.poweroff, battery_critical_low_act);
	lowbat_add_scenario(battery_info.warning, battery_info.realoff, battery_power_off_act);
	lowbat_add_scenario(battery_info.critical, battery_info.critical, battery_critical_low_act);
	lowbat_add_scenario(battery_info.critical, battery_info.realoff, battery_power_off_act);
	lowbat_add_scenario(battery_info.poweroff, battery_info.poweroff, battery_critical_low_act);
	lowbat_add_scenario(battery_info.poweroff, battery_info.realoff, battery_power_off_act);
	lowbat_add_scenario(battery_info.realoff, battery_info.realoff, battery_power_off_act);
}

static void battery_level_send_system_event(int bat_percent)
{
	const char *str;
	static const char *prev;

	if (bat_percent >= battery_info.normal && battery.charge_full == CHARGING_FULL)
		str = EVT_VAL_BATTERY_LEVEL_FULL;
	else if (bat_percent > battery_info.warning)
		str = EVT_VAL_BATTERY_LEVEL_HIGH;
	else if (bat_percent > battery_info.critical)
		str = EVT_VAL_BATTERY_LEVEL_LOW;
	else if (bat_percent > battery_info.poweroff)
		str = EVT_VAL_BATTERY_LEVEL_CRITICAL;
	else
		str = EVT_VAL_BATTERY_LEVEL_EMPTY;

	if (battery_plgn.lowbat_noti_clean && prev)
		battery_plgn.lowbat_noti_clean(prev, str);

	if (prev == str)
		return;

	prev = str;

	_D("System_event: %s", str);
	CRITICAL_LOG("Battery %s.", str);
	event_system_send(SYS_EVENT_BATTERY_LEVEL_STATUS,
			EVT_KEY_BATTERY_LEVEL_STATUS, str);
}

static void change_lowbat_level(int bat_percent)
{
	int prev, now, ret;

	ret = vconf_get_int(VCONFKEY_SYSMAN_BATTERY_LEVEL_STATUS, &prev);
	if (ret < 0) {
		_E("Failed to get vconf value for battery level status: %d", vconf_get_ext_errno());
		return;
	}

	if (bat_percent >= battery_info.normal) {
		if (battery.charge_full == CHARGING_FULL)
			now = VCONFKEY_SYSMAN_BAT_LEVEL_FULL;
		else
			now = VCONFKEY_SYSMAN_BAT_LEVEL_HIGH;
	} else if (bat_percent > battery_info.warning)
		now = VCONFKEY_SYSMAN_BAT_LEVEL_HIGH;
	else if (bat_percent > battery_info.critical)
		now = VCONFKEY_SYSMAN_BAT_LEVEL_LOW;
	else if (bat_percent > battery_info.poweroff)
		now = VCONFKEY_SYSMAN_BAT_LEVEL_CRITICAL;
	else
		now = VCONFKEY_SYSMAN_BAT_LEVEL_EMPTY;

	if (prev != now) {
		ret = vconf_set_int(VCONFKEY_SYSMAN_BATTERY_LEVEL_STATUS, now);
		if (ret < 0)
			_E("Failed to set vconf value for battery level status: %d", vconf_get_ext_errno());
	}
}

static void lowbat_custom_popup(int online, int capacity)
{
	static int custom_type = BATTERY_UNKNOWN;

	if (custom_type == CUSTOM_CONTROL_NONE)
		return;
	if (custom_warning_level == BATTERY_UNKNOWN ||
	    custom_warning_level == battery_info.warning)
		return;

	if (custom_type == BATTERY_UNKNOWN) {
		if (custom_warning_level < battery_info.warning)
			custom_type = LOWER_THAN_WARNING;
		else if (custom_warning_level > battery_info.warning)
			custom_type = HIGHER_THAN_WARNING;
		else {
			custom_type = CUSTOM_CONTROL_NONE;
			_E("Custom type is not defined: %d", custom_warning_level);
			return;
		}
	}

	if (custom_type == LOWER_THAN_WARNING && capacity <= battery_info.critical)
		return;
	if (custom_type == HIGHER_THAN_WARNING && capacity <= battery_info.warning)
		return;

	if (capacity > custom_warning_level) {
		custom_warning_launch = false;
		return;
	}
	if (online == battery.online && custom_warning_launch)
		return;
	if (battery.online > POWER_SUPPLY_TYPE_BATTERY && custom_warning_launch)
		return;

	custom_warning_launch = true;
	_I("Launch custom lowbattery warning popup. online(%d, %d), capacity(%d, %d)", online, battery.online, capacity, custom_warning_level);
	(void)battery_warning_low_act(NULL);
}

void check_extreme_status(int status)
{
	int extreme = 0;
	int ret;

	if (status <= VCONFKEY_SYSMAN_BAT_POWER_OFF ||
	    status > VCONFKEY_SYSMAN_BAT_FULL)
		return;
	ret = vconf_get_int(VCONFKEY_PM_KEY_IGNORE, &extreme);
	if (ret < 0)
		_E("Failed to get vconf value for pm key ignore: %d", vconf_get_ext_errno());
	if (ret != 0 || extreme != 1)
		return;

	ret = vconf_set_int(VCONFKEY_PM_KEY_IGNORE, 0);
	if (ret != 0) {
		_E("Failed to set vconf value for pm key ignore: %d", vconf_get_ext_errno());
		return;
	}
	_I("Release key ignore.");
}

static int lowbat_process(int bat_percent, void *ad)
{
	static int online;
	static bool low_bat_old;
	static int low_bat_skip_cnt = 0;
	int new_bat_capacity;
	int new_bat_state;
	int vconf_state = -1;
	int ret = 0;
	int status = -1;
	bool low_bat = false;
	int result = 0;

	if (battery.online == POWER_SUPPLY_TYPE_UNKNOWN)
		return -EINVAL;

	new_bat_capacity = bat_percent;
	change_lowbat_level(new_bat_capacity);
	lowbat_custom_popup(online, new_bat_capacity);
	battery_level_send_system_event(new_bat_capacity);

	cur_bat_capacity = new_bat_capacity;

	ret = vconf_get_int(VCONFKEY_SYSMAN_BATTERY_STATUS_LOW, &vconf_state);
	if (ret < 0) {
		_E("Failed to get vconf value for battery status low: %d", vconf_get_ext_errno());
		return -EIO;
	}

	if (new_bat_capacity <= battery_info.realoff) {
		if (battery.charge_now == CHARGER_CHARGING || battery.charger_charging == CHARGER_DISABLED) {
			_I("Skip lowbat poweroff during test (c:%d charge:%d online:%d charger_charging:%d)",
				battery.capacity, battery.charge_now, battery.online, battery.charger_charging);
			new_bat_state = battery_info.poweroff;
			if (vconf_state != VCONFKEY_SYSMAN_BAT_POWER_OFF)
				status = VCONFKEY_SYSMAN_BAT_POWER_OFF;
		} else {
			new_bat_state = battery_info.realoff;
			if (vconf_state != VCONFKEY_SYSMAN_BAT_REAL_POWER_OFF)
				status = VCONFKEY_SYSMAN_BAT_REAL_POWER_OFF;
		}
	} else if (new_bat_capacity <= battery_info.poweroff) {
		new_bat_state = battery_info.poweroff;
		if (vconf_state != VCONFKEY_SYSMAN_BAT_POWER_OFF)
			status = VCONFKEY_SYSMAN_BAT_POWER_OFF;
	} else if (new_bat_capacity <= battery_info.critical) {
		new_bat_state = battery_info.critical;
		if (vconf_state != VCONFKEY_SYSMAN_BAT_CRITICAL_LOW)
			status = VCONFKEY_SYSMAN_BAT_CRITICAL_LOW;
	} else if (new_bat_capacity <= battery_info.warning) {
		new_bat_state = battery_info.warning;
		if (vconf_state != VCONFKEY_SYSMAN_BAT_WARNING_LOW)
			status = VCONFKEY_SYSMAN_BAT_WARNING_LOW;
	} else {
		new_bat_state = battery_info.normal;
		if (new_bat_capacity == BATTERY_FULL) {
			if (battery.charge_full) {
				if (vconf_state != VCONFKEY_SYSMAN_BAT_FULL)
					status = VCONFKEY_SYSMAN_BAT_FULL;
			} else {
				if (vconf_state != VCONFKEY_SYSMAN_BAT_NORMAL)
					status = VCONFKEY_SYSMAN_BAT_NORMAL;
			}
		} else {
			if (vconf_state != VCONFKEY_SYSMAN_BAT_NORMAL)
				status = VCONFKEY_SYSMAN_BAT_NORMAL;
		}
	}

	/* If the battery continues to run out even though it is being charged
	 * for a certain period of time, turn off the device. */
	if (new_bat_capacity <= battery_info.realoff && battery.charge_now == CHARGER_CHARGING) {
		if (low_bat_skip_cnt >= RETRY_MAX) {
			new_bat_state = battery_info.realoff;
			status = VCONFKEY_SYSMAN_BAT_REAL_POWER_OFF;
			_I("Go to real poweroff inspite of charging (c:%d charge:%d online:%d charger_charging %d current now %d)",
				battery.capacity, battery.charge_now, battery.online, battery.charger_charging, battery.current_now);
			low_bat_skip_cnt = 0;
		} else if (battery.current_now <= MIN_INOW_VALUE) {
			low_bat_skip_cnt++;
		} else {
			low_bat_skip_cnt = 0;
		}
	} else {
		low_bat_skip_cnt = 0;
	}

	if (status != -1) {
		result = status;
		if (disp_plgn.update_pm_setting)
			disp_plgn.update_pm_setting(SETTING_LOW_BATT, status);
	} else
		result = vconf_state;

	check_extreme_status(status);

	if (new_bat_capacity <= battery_info.warning)
		low_bat = true;

	if (low_bat_old != low_bat) {
		device_notify(DEVICE_NOTIFIER_LOWBAT, (void *)&low_bat);
		low_bat_old = low_bat;
	}

	if (battery.online == POWER_SUPPLY_TYPE_UNKNOWN)
		return result;
	if (cur_bat_state == new_bat_state && online == battery.online)
		return result;

	online = battery.online;
	if (cur_bat_state == BATTERY_UNKNOWN)
		cur_bat_state = battery_info.normal;
	if (lowbat_scenario(cur_bat_state, new_bat_state, NULL))
		_I("Cur(%d) new(%d) capacity(%d).", cur_bat_state, new_bat_state, bat_percent);
		cur_bat_state = new_bat_state;

	return result;
}

static int check_lowbat_percent(int *pct)
{
	int bat_percent;

	bat_percent = battery.capacity;
	if (bat_percent < 0)
		return -ENODEV;
	if (bat_percent > 100)
		bat_percent = 100;
	change_lowbat_level(bat_percent);
	battery_level_send_system_event(bat_percent);
	*pct = bat_percent;
	return 0;
}

static int lowbat_monitor(void *data)
{
	int bat_percent, r;

	r = lowbat_initialized(NULL);
	if (r != BATTERY_ENABLED)
		return battery.charging_level;

	if (data == NULL) {
		r = check_lowbat_percent(&bat_percent);
		if (r < 0)
			return battery.charging_level;
	} else
		bat_percent = *(int *)data;
	return lowbat_process(bat_percent, NULL);
}

static int lowbat_monitor_init(void *data)
{
	int status;

	status = lowbat_initialized(NULL);
	if (status != BATTERY_UNKNOWN)
		return 0;

	status = BATTERY_ENABLED;
	lowbat_initialized(&status);

	/* it's called just this once. */
	unregister_notifier(DEVICE_NOTIFIER_POWER_SUPPLY, lowbat_monitor_init);

	/* load battery configuration file */
	battery_config_load(&battery_info);
	_I("Battery conf: %d %d %d %d %d", battery_info.normal, battery_info.warning,
		battery_info.critical, battery_info.poweroff, battery_info.realoff);

	lowbat_scenario_init();
	check_lowbat_percent(&battery.capacity);
	lowbat_process(battery.capacity, NULL);
	return 0;
}


static GVariant * dbus_set_lowbat_level(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret = 0;
	int now;
	int status;

	g_variant_get(param, "(i)", &now);

	status = lowbat_initialized(NULL);
	if (status != BATTERY_ENABLED) {
		_E("Invalid lowbat handler.");
		ret = -EINVAL;
		goto out;
	}

	if (now <= battery_info.warning && now != BATTERY_UNKNOWN) {
		_E("Invalid level is requested: %d", now);
		ret = -EINVAL;
		goto out;
	}

	custom_warning_level = now;
	custom_warning_launch = false;

	ret = vconf_set_int(VCONF_KEY_BATTERY_WARNING_LEVEL, custom_warning_level);
	if (ret < 0)
		_E("Failed to set vconf value for battery warning level: %d", vconf_get_ext_errno());
	_D("Custom warning level is changed to %d.", custom_warning_level);

out:
	return g_variant_new("(i)", ret);
}

static GVariant * dbus_get_lowbat_level(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	_D("Warning level: %d", custom_warning_level);
	return g_variant_new("(i)", custom_warning_level);
}

static const dbus_method_s dbus_methods[] = {
	{ "SetLowbatLevel",  "i", "i", dbus_set_lowbat_level },
	{ "GetLowbatLevel",  NULL, "i", dbus_get_lowbat_level },
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_BATTERY,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static int lowbat_probe(void *data)
{
	static const struct device_ops *ops;
	int ret = -EINVAL;

	FIND_DEVICE_INT(ops, "power_supply");
	ret = ops->probe(data);
	if (ret == 0)
		_I("Support lowbat handler.");

	return ret;
}

static void lowbat_init(void *data)
{
	int ret;

	register_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
	register_notifier(DEVICE_NOTIFIER_POWER_SUPPLY, lowbat_monitor_init);

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_BATTERY, &dbus_interface);
	if (ret < 0)
		_E("Failed to init dbus method: %d", ret);
}

static void lowbat_exit(void *data)
{
	int status = BATTERY_DISABLED;

	lowbat_initialized(&status);
	if (low_batt_sig_timer) {
		g_source_remove(low_batt_sig_timer);
		low_batt_sig_timer = 0;
	}
}

static gboolean low_battery_charge_status(void *data)
{
	low_batt_sig_timer = 0;
	battery.charging_level = lowbat_monitor(data);
	if (battery.charging_level > 0 && old_battery.charging_level != battery.charging_level) {
		if (vconf_set_int(VCONFKEY_SYSMAN_BATTERY_STATUS_LOW, battery.charging_level) < 0)
			_E("Fail to set level.");
		if (power_supply_broadcast(CHARGE_LEVEL_SIGNAL, battery.charging_level) < 0)
			_E("power_supply_broadcast failed");
	}
	return G_SOURCE_REMOVE;
}

static int lowbat_execute(void *data)
{
	/* In battery kernel side, it has a requirement that battery charging event must get quickly.
	 * So deviced receives the discharging and charging event within 1.5 sec,
	 * added a timer to wait for the second signal, if there is no second signal then execute
	 * the lowbat_execute.
	 */
	if (low_batt_sig_timer)
		g_source_remove(low_batt_sig_timer);

	low_batt_sig_timer = g_timeout_add(1500, low_battery_charge_status, data);

	return 0;
}

static const struct device_ops lowbat_device_ops = {
	DECLARE_NAME_LEN("lowbat"),
	.probe    = lowbat_probe,
	.init     = lowbat_init,
	.execute  = lowbat_execute,
	.exit     = lowbat_exit,
};

DEVICE_OPS_REGISTER(&lowbat_device_ops)
