/*
 * devicectl
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <libsyscommon/dbus-system.h>
#include <core/common.h>
#include "usb.h"

/*
 * devicectl [device] [action]
 * ex> devicectl display stop
 *     devicectl pass start
 */

enum device_type {
	DEVICE_CORE,
	DEVICE_DISPLAY,
	DEVICE_LED,
	DEVICE_PASS,
	DEVICE_USB,
	DEVICE_EXTCON,
	DEVICE_POWER,
	DEVICE_USB_HOST_TEST,
	DEVICE_MAX,
	DEVICE_ALL,
};
static enum device_type arg_id;

static const struct device {
	const enum device_type id;
	const char *name;
	const char *path;
	const char *iface;
} devices[] = {
	{ DEVICE_CORE,    "core",    DEVICED_PATH_CORE,    DEVICED_INTERFACE_CORE    },
	{ DEVICE_DISPLAY, "display", DEVICED_PATH_DISPLAY, DEVICED_INTERFACE_DISPLAY },
	{ DEVICE_LED,     "led",     DEVICED_PATH_LED,     DEVICED_INTERFACE_LED     },
	{ DEVICE_PASS,    "pass",    DEVICED_PATH_PASS,    DEVICED_INTERFACE_PASS    },
	{ DEVICE_USB,     "usb",     DEVICED_PATH_USB,     DEVICED_INTERFACE_USB     },
	{ DEVICE_EXTCON,  "extcon",  DEVICED_PATH_EXTCON,  DEVICED_INTERFACE_EXTCON  },
	{ DEVICE_POWER,   "power",   DEVICED_PATH_POWEROFF,  DEVICED_INTERFACE_POWEROFF  },
	{ DEVICE_USB_HOST_TEST, "usb-host-test", DEVICED_PATH_USB_HOST_TEST, DEVICED_INTERFACE_USB_HOST_TEST},
};

static int start_device(char **args)
{
	GVariant *msg;

	if (!args[1])
		return -EINVAL;

	printf("Start %s device.\n", args[1]);

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
		devices[arg_id].path,
		devices[arg_id].iface,
		"start",
		NULL);
	if (!msg)
		return -EBADMSG;

	g_variant_unref(msg);

	return 0;
}

static int stop_device(char **args)
{
	GVariant *msg;

	if (!args[1])
		return -EINVAL;

	printf("Stop %s device.\n", args[1]);

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
						devices[arg_id].path,
						devices[arg_id].iface,
						"stop",
						NULL);
	if (!msg)
		return -EBADMSG;

	g_variant_unref(msg);

	return 0;
}

static int dump_mode(char **args)
{
	GVariant *msg;
	int ret = -EBADMSG;

	if (!args[1] || !args[2] || !args[3]) {
		ret = -EINVAL;
		goto out;
	}

	printf("%s (%s %s).\n", args[1], args[2], args[3]);

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
					devices[arg_id].path,
					devices[arg_id].iface,
					"Dumpmode",
					g_variant_new("(s)", args[3]));
	if (!msg) {
		printf("Failed to set dump mode");
		ret = -EBADMSG;
		goto out;
	}

	if (!dh_get_param_from_var(msg, "(i)", &ret)) {
		printf("Failed to get signature(%s): no message", g_variant_get_type_string(msg));
		ret = -EBADMSG;
	}

	g_variant_unref(msg);

out:
	return ret;
}

static int display_state(char **args)
{
	GVariant *msg;
	int ret = -EBADMSG;

	if (!args[1] || !args[2] || !args[3]) {
		ret = -EINVAL;
		goto out;
	}

	printf("%s (%s %s)!\n", args[1], args[2], args[3]);

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
		    devices[arg_id].path,
		    devices[arg_id].iface,
		    "changestate",
		    g_variant_new("(s)", args[3]));

	if (!msg) {
		printf("Failed to set display state");
		ret = -EBADMSG;
		goto out;
	}

	if (!dh_get_param_from_var(msg, "(i)", &ret)) {
		printf("Failed to get signature(%s): no message", g_variant_get_type_string(msg));
		ret = -EBADMSG;
	}

	g_variant_unref(msg);

out:
	return ret;
}

static int save_log(char **args)
{
	GVariant *msg;

	if (!args[1])
		return -EINVAL;

	printf("Save log %s device.\n", args[1]);

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
				devices[arg_id].path,
				devices[arg_id].iface,
				"SaveLog",
				NULL);
	if (!msg) {
		printf("Failed to save log");
		return -EBADMSG;
	}

	g_variant_unref(msg);

	return 0;
}

static void get_pname(pid_t pid, char *pname)
{
	char buf[PATH_MAX];
	int cmdline, r;

	snprintf(buf, PATH_MAX, "/proc/%d/cmdline", pid);
	cmdline = open(buf, O_RDONLY);
	if (cmdline < 0) {
		pname[0] = '\0';
		return;
	}

	r = read(cmdline, pname, PATH_MAX);
	if ((r >= 0) && (r < PATH_MAX))
		pname[r] = '\0';
	else
		pname[0] = '\0';

	close(cmdline);
}

static int save_dbus_name(char **args)
{
	GVariant *msg;
	GVariantIter *iter;
	char *str;
	pid_t pid;
	char pname[PATH_MAX];

	if (!args[1])
		return -EINVAL;

	printf("Save dbus name.\n");

	msg = dbus_handle_method_sync_with_reply_var(DBUS_BUS_NAME,
						DBUS_OBJECT_PATH,
						DBUS_INTERFACE_NAME,
						"ListNames",
						NULL);
	if (!msg) {
		printf("Failed to get list names.");
		return -EBADMSG;
	}

	if (!dh_get_param_from_var(msg, "(as)", &iter)) {
		printf("Invalid list name arguments.");
		g_variant_unref(msg);
		return -EINVAL;
	}

	printf("%" G_GSIZE_FORMAT " connections.\n", g_variant_iter_n_children(iter));

	while (g_variant_iter_loop(iter, "s", &str)) {
		pid = dbus_handle_get_sender_pid(NULL, str);
		get_pname(pid, pname);
		printf("%6d  %6s  %s\n", pid, str, pname);
	}

	g_variant_iter_free(iter);
	g_variant_unref(msg);

	return 0;
}

static int device_list(char **args)
{
	GVariant *msg;

	if (!args[1])
		return -EINVAL;

	printf("Print %s to dlog.\n", args[1]);

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
						devices[arg_id].path,
						devices[arg_id].iface,
						"DeviceList",
						NULL);
	if (!msg)
		return -EBADMSG;

	g_variant_unref(msg);

	return 0;
}

static int set_usb_mode(char **args)
{
	return load_usb_mode(args[3]);
}

static int unset_usb_mode(char **args)
{
	return unload_usb_mode(args[3]);
}

static int enable_device(char **args)
{
	GVariant *msg;
	int ret = -EBADMSG;

	if (!args[3]) {
		ret = -EINVAL;
		goto out;
	}

	printf("Enable %s device.\n", args[3]);

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
						devices[arg_id].path,
						devices[arg_id].iface,
						"enable",
						g_variant_new("(s)", args[3]));
	if (!msg) {
		ret = -EBADMSG;
		goto out;
	}

	if (!dh_get_param_from_var(msg, "(i)", &ret)) {
		printf("Failed to get signature(%s): no message", g_variant_get_type_string(msg));
		ret = -EBADMSG;
	}

	g_variant_unref(msg);

out:
	return ret;
}

static int disable_device(char **args)
{
	GVariant *msg;
	int ret = -EBADMSG;

	if (!args[3]) {
		ret = -EINVAL;
		goto out;
	}

	printf("Disable %s device.\n", args[3]);

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
						devices[arg_id].path,
						devices[arg_id].iface,
						"disable",
						g_variant_new("(s)", args[3]));
	if (!msg) {
		ret = -EBADMSG;
		goto out;
	}

	if (!dh_get_param_from_var(msg, "(i)", &ret)) {
		printf("Failed to get signature(%s): no message", g_variant_get_type_string(msg));
		ret = -EBADMSG;
	}

	g_variant_unref(msg);

out:
	return ret;
}

static int power_operation(char **args, char *type)
{
	GVariant *msg;
	int ret = -EBADMSG;

	if (!args[1] || !type) {
		ret = -EINVAL;
		goto out;
	}

	printf("Power %s device.\n", args[2]);

	msg = dbus_handle_method_sync_with_reply_var(DEVICED_BUS_NAME,
						devices[arg_id].path,
						devices[arg_id].iface,
						"PowerOff",
						g_variant_new("(s)", type));
	if (!msg) {
		ret = -EBADMSG;
		goto out;
	}

	if (!dh_get_param_from_var(msg, "(i)", &ret)) {
		printf("Failed to get signature(%s): no message", g_variant_get_type_string(msg));
		ret = -EBADMSG;
	}

	g_variant_unref(msg);

out:
	return ret;
}

static int power_off(char **args)
{
	return power_operation(args, "poweroff");
}

static int power_reboot(char **args)
{
	return power_operation(args, "reboot");
}

static const struct action {
	const enum device_type id;
	const char *action;
	const int argc;
	int (* const func)(char **args);
	const char *option;
} actions[] = {
	{ DEVICE_ALL,       "start",           3, start_device,      ""            },
	{ DEVICE_ALL,       "stop",            3, stop_device,       ""            },
	{ DEVICE_DISPLAY,   "dumpmode",        4, dump_mode,         "[on|off]"    },
	{ DEVICE_DISPLAY,   "state",           4, display_state,     "[lcdon|lcdoff]" },
	{ DEVICE_LED,       "dumpmode",        4, dump_mode,         "[on|off]"    },
	{ DEVICE_DISPLAY,   "savelog",         3, save_log,          ""            },
	{ DEVICE_USB,       "set",             4, set_usb_mode,      "[sdb|ssh]"   },
	{ DEVICE_USB,       "unset",           4, unset_usb_mode,    "[sdb|ssh]"   },
	{ DEVICE_CORE,      "dbusname",        3, save_dbus_name,    ""            },
	{ DEVICE_CORE,      "devicelist",      3, device_list,       ""            },
	{ DEVICE_EXTCON,    "enable",          4, enable_device,     "[USB|HEADPHONE|HDMI|DOCK]" },
	{ DEVICE_EXTCON,    "disable",         4, disable_device,    "[USB|HEADPHONE|HDMI|DOCK]" },
	{ DEVICE_POWER,     "off",             3, power_off,         ""            },
	{ DEVICE_POWER,     "reboot",          3, power_reboot,      ""            },
};

static inline void usage()
{
	printf("[usage] devicectl <device_name> <action>\n");
	printf("Please use option --help to check options.\n");
}

static void help()
{
	int i;

	printf("[usage] devicectl <device_name> <action> <option>\n");
	printf("device name & action & option\n");
	for (i = 0; i < ARRAY_SIZE(actions); i++) {
		if (actions[i].id == DEVICE_ALL) {
			printf("    [all-device] %s %s\n", actions[i].action,
			    actions[i].option);
		} else {
			printf("    %s %s %s\n", devices[actions[i].id].name,
			    actions[i].action, actions[i].option);
		}
	}
}

int main(int argc, char *argv[])
{
	int i;

	if (argc == 2 && !strcmp(argv[1], "--help")) {
		help();
		return 0;
	}

	if (argc < 3) {
		usage();
		return -EINVAL;
	}

	for (i = 0; i < argc; i++)
		if (argv[i] == NULL) {
			usage();
			return -EINVAL;
		}

	for (i = 0; i < ARRAY_SIZE(devices); i++)
		if (!strcmp(argv[1], devices[i].name))
			break;

	if (i >= ARRAY_SIZE(devices)) {
		printf("Invalid device name. %s\n", argv[1]);
		usage();
		return -EINVAL;
	}

	arg_id = devices[i].id;

	for (i = 0; i < ARRAY_SIZE(actions); i++)
		if (actions[i].id == arg_id || actions[i].id == DEVICE_ALL)
			if (!strcmp(argv[2], actions[i].action))
				break;

	if (i >= ARRAY_SIZE(actions)) {
		printf("Invalid action name. %s\n", argv[2]);
		usage();
		return -EINVAL;
	}

	if (actions[i].argc != argc) {
		printf("Invalid arg count.\n");
		usage();
		return -EINVAL;
	}

	return actions[i].func(argv);
}

