/*
 * deviced
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdbool.h>

#include <vconf.h>
#include <hw/usb_gadget.h>

#include "core/log.h"
#include "core/device-notifier.h"

#include "usb.h"

static bool debug_state;

int get_usb_debug_state(void)
{
	int state;

	/* Never use debug_state variable, anyone other than deviced can change this vconf. */

	if (vconf_get_bool(VCONFKEY_SETAPPL_USB_DEBUG_MODE_BOOL, &state) != VCONF_OK) {
		_E("Failed to get vconf value for USB debug mode: %d", vconf_get_ext_errno());
		return -1;
	}

	return state;
}

void set_usb_debug_state(bool on)
{
	int state = on ? 1 : 0;

	/*
	 * VCONFKEY_SETAPPL_USB_DEBUG_MODE_BOOL can also be written outside of deviced.
	 * So you should always write regardless of the internal debug_state of deviced.
	 */
	if (vconf_set_bool(VCONFKEY_SETAPPL_USB_DEBUG_MODE_BOOL, state) == VCONF_OK)
		debug_state = on;
	else
		_E("Failed to set vconf value for USB debug mode: %d", vconf_get_ext_errno());
}

static void usb_debug_changed(keynode_t *key, void *data)
{
	int mode;

	if (!key)
		return;

	mode = vconf_keynode_get_bool(key);

	/* This also prevents vconf self-loop by set_usb_debug_state(). */
	if (mode == debug_state) {
		_I("USB debug mode is already %s.", mode ? "ON" : "OFF");
		return;
	}

	debug_state = mode;

	_I("USB debug mode is changed to %s.", mode ? "ON" : "OFF");

	device_notify(DEVICE_NOTIFIER_USB_DEBUG_MODE, (void *)mode);
}

static int usb_debug_mode_changed(void *on)
{
	int ret;
	unsigned int new_mode;
	bool rndis_is_enabled = false;
	const unsigned int curr_mode = usb_state_get_selected_mode();

	if (curr_mode & USB_FUNCTION_RNDIS)
		rndis_is_enabled = true;

	/*
	 * debug  on + rndis  on : USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_RNDIS
	 * debug  on + rndis off : USB_FUNCTION_MTP | USB_FUNCTION_ACM | USB_FUNCTION_SDB
	 * debug off + rndis  on : USB_FUNCTION_RNDIS
	 * debug off + rndis off : USB_FUNCTION_MTP | USB_FUNCTION_ACM
	 */

	if ((int)on) {
		if (rndis_is_enabled)
			new_mode = USB_FUNCTION_ACM | USB_FUNCTION_SDB | USB_FUNCTION_RNDIS;
		else
			new_mode = USB_FUNCTION_MTP | USB_FUNCTION_ACM | USB_FUNCTION_SDB;
	} else {
		if (rndis_is_enabled)
			new_mode = USB_FUNCTION_RNDIS;
		else
			new_mode = USB_FUNCTION_MTP | USB_FUNCTION_ACM;
	}

	ret = usb_change_mode(new_mode, false);
	if (ret < 0)
		_E("Failed to change USB mode to (%d).", new_mode);

	return ret;
}

void add_usb_debug_handler(void)
{
	int state = 0;

	if (vconf_notify_key_changed(VCONFKEY_SETAPPL_USB_DEBUG_MODE_BOOL,
				usb_debug_changed, NULL) != VCONF_OK)
		_E("Failed to add USB debug handler.");

	register_notifier(DEVICE_NOTIFIER_USB_DEBUG_MODE, usb_debug_mode_changed);

	if (vconf_get_bool(VCONFKEY_SETAPPL_USB_DEBUG_MODE_BOOL, &state) == VCONF_OK)
		debug_state = (state == 0 ? false : true);
	else
		_E("Failed to get vconf value for USB debug mode: %d", vconf_get_ext_errno());
}

void remove_usb_debug_handler(void)
{
	vconf_ignore_key_changed(VCONFKEY_SETAPPL_USB_DEBUG_MODE_BOOL, usb_debug_changed);
	unregister_notifier(DEVICE_NOTIFIER_USB_DEBUG_MODE, usb_debug_mode_changed);
}
